package xmdv.dataproc;

/***********************************************************
**  CopyRight (c) by Xmdv Group, CS Department of WPI, 2003
**  All Rights Reserved.
**
**  Author: Jing Yang
**  Date:   06/2003
***********************************************************/

import java.util.*;
import java.io.*;
import xmdv.graphics.*;

/**
 * This class defines a dataset. 
 * It contains the dimension (variable) 
 * and brushing information and the data items. 
 *
 * @author  Jing Yang
 * @version v1.0
 */
public class Dataset {
    // Variables declaration
    
    /** number of data points in this dataset, i.e., the cardinality of this dataset */
    protected int dataNum;
    
    /** number of dimensions of this dataset, all dimensions are counted, */ 
    protected int allDimNum;
    
    /** all dimensoins of this dataset */
    protected AbstractDimension[] allDims = null;
    
    /** a data buffer tht contains the normalized value of all data points in this dataset */
    protected double[][] norDataBuf = null;
    
    /** number of dimensions that are currently in use */
    protected int currDimNum;
    
    /** dimensions that are currently in use */
    protected AbstractDimension[] currDims = null;
    
    /** current dimension index list
     * an arraylist that contains the indexes of the dimensions that are currently in use. 
     * The index of a dimension is decided by its order in all dimenions. 
     * ([0,N-1] in a N dimensional dataset)
     */
    protected ArrayList currDimIndexList = null;
    
    /** current dimension indexes */
    protected int[] currDimIndexes = null;
    
    /** a hash set that contains all the canvases that linked to this dataset */
    protected HashSet canvases = null;
    
    /** all Brush (Brush on all dimensions) */
    protected Brush allBrush = null;
    
    /** current Brush */
    protected Brush currBrush = null;
    
    /** BrushOper */
    protected BrushOper brushOper = null;
    
    /** marks of data points */
    protected int[] brushMarks = null;
    
    /** multi-selection set*/
    protected HashSet multiSelects = null;
    
    /** multi-selection state */
    protected boolean multiSelectIsOn = false;
    
    /** delimiters in the input file*/
    private String delimiters = " ,;\t\n\r";
    // End of variables declaration
    
    
    /** Creates new dataset*/
    public Dataset() {
        canvases = new HashSet();
    }
    
    /** Gets multi-selection state 
     * @return multi-selection state 
     */ 
    public boolean isMultiSelectOn() {
        return multiSelectIsOn;
    }
    
    /** Sets multi-selection state 
     * @param state true to clear multi-selection set, 
     * according to current multi-selection,
     * false to reset current Brush turn it off
     */
    public void multiSelectSetState(boolean state)
    {
        multiSelectIsOn = state;
    
        if (state) {
            multiSelects = new HashSet();
        } else {
            if (multiSelects.size() != 0) {
                multiSelectDefineBrush();
            }
        }
    }
    
    /** 
     * Adds a data point to the multi-selection set 
     * @param index the index of the data point to be added
     * @return false if the data point already exists in the multi-selection set, else returns true 
     */
    public boolean multiSelectAddPoint(int index) {
        return multiSelects.add(new Integer(index));
    }
    
    /** 
     * Removes a data point from the multi-selection set 
     * @param index the index of the data point to be removed
     * @return false if the data point does not exist in the multi-selection set, else returns true 
     */
    public boolean multiSelectRemovePoint(int index) {
        return multiSelects.remove(new Integer(index));
    }
    
    
    /** 
     * Judges if a data point is contained in the multi-selection set 
     * @param index the index of the data point to be checked
     * @return true if the data point exists in the multi-selection set, else returns false
     */
    public boolean multiSelectContains(int index) {
        return multiSelects.contains(new Integer(index));
    }
    
    
    /** Sets current Brush so that it exactly covers the data points 
     * in the multi-selection set
     */
    public void multiSelectDefineBrush() {
        int i, j;
        double[] mins = new double[currDimNum];
        double[] maxs = new double[currDimNum];
        // Set starting value for mins/maxes
        for (i = 0; i < currDimNum; i++)
        {
            mins[i] = 1;
            maxs[i] = 0;
        }
        
        Iterator it = multiSelects.iterator();
        
        while (it.hasNext()) {
            int index = ((Integer)it.next()).intValue();
            double[] data = getCurrNorData(index);
            for (i = 0; i < currDimNum; i++)
            {
                if (data[i] < mins[i])
                   mins[i] = data[i];
                if (data[i] > maxs[i])
                   maxs[i] = data[i];
            }
        }
    
        for (i = 0; i < currDimNum; i++) {
            currBrush.setNorMinMax(i, mins[i], maxs[i]);
        }
    
        brush();
    }
    
    /** Add a related canvas to this dataset
     * @param canvas the canvas to be added
     */
    public void addCanvas(grCanvas canvas) {
        canvases.add(canvas);
    }
    
    /** Removes a related canvas of this dataset
     * @param canvas the canvas to be removed
     */
    public void removeCanvas(grCanvas canvas) {
        canvases.remove(canvas);
    }
    
    /** Refreshes all canvases related to this dataset
     */
    public void refreshAllCanvas() {
        Iterator ite = canvases.iterator();
        while (ite.hasNext()) {
            ((grCanvas)ite.next()).refresh();
        }
    }    
    
    /** Gets the number of current dimensions
     * @return the number of current dimensions
     */
    public int getCurrDimNum() {
        return currDimNum;
    }
    
    /** Gets the number of data points in the dataset
     * @return the number of data points in the dataset
     */
    public int getDataNum() {
        return dataNum;
    }
    
    /** Gets current dimensions
     * @return current dimensions
     */
    public AbstractDimension[] getCurrDims() {
        return currDims;
    }
    
    /** Gets current Brush 
     * @return current Brush
     */
    public Brush getCurrBrush() {
        return currBrush;
    }
    
    /** Gets the normalized value of a data point in current dimensions
     * @param dataIndex the index of the data point
     * @return the normalized value of a data point in current dimensions
     */
    public double[] getCurrNorData(int dataIndex) {
        double[] currNorData = new double[currDimNum];
        for (int i = 0; i < currDimNum; i++ ) {
            currNorData[i] = norDataBuf[dataIndex][currDimIndexes[i]];
        }
        return currNorData;
    }
    
    /** Cascades the change of current dimension index list 
     * This method should be called after the current dimension index list is changed.
     */
    public void onCurrDimIndexListChanged() {
        //update current dimension indexes
        currDimNum = currDimIndexList.size();
        currDimIndexes = new int[currDimNum];
    
        currDims = new AbstractDimension[currDimNum];
        currBrush = new Brush(currDimNum);
        
        int index, i;
        for (i = 0; i < currDimNum; i++ ) {
            index = ((Integer)currDimIndexList.get(i)).intValue();
            currDimIndexes[i] = index;
            currDims[i] = allDims[index];
            currBrush.setBrush1D(i, allBrush.getBrush1D(index));
        }
        
        //update marks of data points
        brush();
        
        //update related canvases
        Iterator ite = canvases.iterator();
        while (ite.hasNext()) {
            GeneralDisplay display = ((grCanvas)ite.next()).getDisplay();
            display.updateParameters();
        }
    }    
    
    /** Sets the BrushOper of this dataset
     * @param brushOper the BrushOper to be set
     */
    public void setBrushOper(BrushOper brushOper) {
        this.brushOper = brushOper;
    }
    
    /** Gets the BrushOper of this dataset
     * @return the BrushOper of this dataset
     */
    public BrushOper getBrushOper() {
        return brushOper;
    }
    
    /** Loads data from a file
     * @param fileName the name of the file
     * @return true if the load suceeds. Returns false if it fails.
     */
    public boolean load(String fileName) {
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            readMetaData(reader);
            readData(reader);
            reader.close();
    
        } catch (FileNotFoundException e){
            return false;
        } catch (IOException e2){
            return false;
        }
        
        initAllBrush();
        initBrushOperAndMarks();
        initCurrDimIndexList();
        return true;
    }
    
    /** Reads meta data of the input file */
    protected void readMetaData(BufferedReader reader) throws IOException {
        String line;
        StringTokenizer tokens;
        
        //read number of dimensions and number of data items
        line = reader.readLine();
        tokens = new StringTokenizer(line, delimiters);
        allDimNum = Integer.parseInt(tokens.nextToken());
        dataNum = Integer.parseInt(tokens.nextToken());
    
        //new dimensions
        allDims = new AbstractDimension[allDimNum];
        
        //read name of dimensions
        for ( int i = 0; i < allDimNum; i++ ) {
            allDims[i] = new DoubleDimension();
            allDims[i].setName(reader.readLine());
        }
        
        //read dimension minimum, maximum, and dimension cardinality for dimension stacking
        for ( int i = 0; i < allDimNum; i++ ) {
            line = reader.readLine();
    
            tokens = new StringTokenizer(line, delimiters);
            allDims[i].setMin(tokens.nextToken());
            allDims[i].setMax(tokens.nextToken());
            allDims[i].setCardinality(Integer.parseInt(tokens.nextToken()));
            allDims[i].addRim();
        }
    }
        
    /** Reads tuples of the input file */    
    protected void readData(BufferedReader reader) throws IOException {
        int i, j;
    
        norDataBuf = new double[dataNum][allDimNum];
        for (i = 0; i < dataNum; i ++) {
            String line;
            line = reader.readLine();
            StringTokenizer tokens = new StringTokenizer(line, delimiters);
            for (j = 0; j < allDimNum; j++) {
                norDataBuf[i][j] = allDims[j].getNorValue(tokens.nextToken());
                    //Double.parseDouble(tokens.nextToken());
            }
        }
    }
       
    /** Resets the marks of all data points
     * the marks of all the data points will be the clear mark of the BrushOper of this dataset
     */     
    public void resetBrushMarks() {
        int mark = brushOper.getClearMark();
        for (int i = 0; i < dataNum; i++ ) {
            brushMarks[i] = mark;
        }
    }
    
    /** Replaces the marks of the data points with a certain mark by another mark
     * @param oldMark the old mark that needs to be replaced
     * @param newMark the new mark that will replace the old mark 
     */
    public void replaceMarks(int oldMark, int newMark) {
        for (int i = 0; i < dataNum; i++ ) {
            if (brushMarks[i] == oldMark) {
                brushMarks[i] = newMark;
            }
        }
    }
       
    /** init the all Brush */   
    protected void initAllBrush() {
        allBrush = new Brush(allDimNum);
    }
    
    /** inital BrushOper and marks of all data points */
    protected void initBrushOperAndMarks() {
        brushOper = new BrushOper(BrushOper.REGULAR, BrushOper.HIGHLIGHTED);
        brushMarks = new int[dataNum];
        resetBrushMarks();
    }
    
    /** initial current dimension index list */
    protected void initCurrDimIndexList() {
        int i;
        
        currDimIndexList = new ArrayList();
        
        for (i = 0; i < allDimNum; i++) {
            currDimIndexList.add(new Integer(i));
        }
        
        onCurrDimIndexListChanged();
    }
    
    /** Brushes the dataset 
     * if the current value of a data point is within current Brush, 
     * then this data point will be marked current mark of the BrushOper of the dataset
     * elsewise it will be marked clear mark of the BrushOper.
     */
    public void brush() {
        double[] norData;
        int clearMark = brushOper.getClearMark();
        int currMark = brushOper.getCurrMark();
        
        for (int i = 0; i < dataNum; i++ )
        {
            if (brushMarks[i] == currMark)
                brushMarks[i] = clearMark;
                
            norData = getCurrNorData(i);
            if (currBrush.isDataWithinBrush(norData))
                brushMarks[i] = currMark;
        }
    }
    
    /** Gets the mark of a data point
     * @param dataIndex the index of the data point 
     * @return the mark of the data point 
     */
    public int getMark(int dataIndex) {
        if (dataIndex < 0 || dataIndex > dataNum - 1)
            return -1;
        return brushMarks[dataIndex];
    }
}
